// Clearing.cpp : main project file.

#include "stdafx.h"

#define PRINT_LOCAL_INFO Print Info to stderr with cerr
#undef PRINT_LOCAL_INFO

const __int64 FIRST_RUN_TIME = 20000;
const __int64 SECOND_RUN_TIME = 32000;

const char * INPUT_FILE = "cleaning.in";
const char * OUTPUT_FILE = "cleaning.out";


Timer * timer = new Timer(FIRST_RUN_TIME, SECOND_RUN_TIME);

Input input(INPUT_FILE);
Solver solver(timer);
Answer answer, currentAnswer;



void PrintAnswer()
{
#ifdef PRINT_LOCAL_INFO
	cerr << "\n\nPrinting at RunTime: " << timer -> GetElapsedTime() << "\n";
#endif

	answer.saveAnswerToFile(OUTPUT_FILE);

	int leastAnswer = input.GetTotalArea() / (input.GetBoxWidth() * input.GetBoxHeight());
	if(input.GetTotalArea() % (input.GetBoxWidth() * input.GetBoxHeight()) > 0)
		leastAnswer ++;

#ifdef PRINT_LOCAL_INFO
	cerr << "Answer Bins Used: " << answer.GetNumberOfBins() << "\n";
	cerr << "Least Bins Used: " << leastAnswer << "\n";
	cerr << "Least Score: " << pow(double(leastAnswer)/double(answer.GetNumberOfBins()), double(3.0)) << "\n";
	cerr << "PROGRAM RUN TIME: " << timer -> GetElapsedTime() << "\n";
	cerr << "CLOCK RUN TIME: " << double(clock()) / double(CLOCKS_PER_SEC) << "\n";
	printf("%.9lf %.9lf\n", pow(double(leastAnswer)/double(answer.GetNumberOfBins()), double(3.0)), timer -> GetElapsedTime());
#endif

	//cerr << "PROGRAM RUN TIME: " << timer -> GetElapsedTime() << "\n";
	//cerr << "CLOCK RUN TIME: " << double(clock()) / double(CLOCKS_PER_SEC) << "\n";

	//system("pause");
}

void FirstSolvers()
{
	timer -> NextState();
#ifdef PRINT_LOCAL_INFO
	cerr << "\n\nStarting First Solvers at RunTime: " << timer -> GetElapsedTime() << "\n";
#endif
	
	
	// Global
	answer.CheckMin(solver.GuillotineGlobal(input, false, GuillotineBinPack::FreeRectChoiceHeuristic(0), GuillotineBinPack::GuillotineSplitHeuristic(2)));
	if(! timer -> TimeLeft()) return;
#ifdef PRINT_LOCAL_INFO
	if(timer -> TimeLeft())
		cerr << "GlobalGuillotineGlobal(false,0,2) Complete: " << answer.GetNumberOfBins() << "\n";
#endif

	input.SortDecreasingArea();
	answer.CheckMin(solver.MaxRects(input, MaxRectsBinPack::FreeRectChoiceHeuristic(0)));
	if(! timer -> TimeLeft()) return;
#ifdef PRINT_LOCAL_INFO
	cerr << "SortDecreasingAreaMaxRects(0) Complete: " << answer.GetNumberOfBins() << "\n";
#endif
	answer.CheckMin(solver.MaxRects(input, MaxRectsBinPack::FreeRectChoiceHeuristic(2)));
	if(! timer -> TimeLeft()) return;
#ifdef PRINT_LOCAL_INFO
	cerr << "SortDecreasingAreaMaxRects(2) Complete: " << answer.GetNumberOfBins() << "\n";
#endif


	// Global
	answer.CheckMin(solver.MaxRectsGlobal(input, MaxRectsBinPack::FreeRectChoiceHeuristic(2)));
	if(! timer -> TimeLeft()) return;
#ifdef PRINT_LOCAL_INFO
	cerr << "GlobalMaxRectsGlobal(2) Complete: " << answer.GetNumberOfBins() << "\n";
#endif


	input.SortDecreasingSide();

	answer.CheckMin(solver.MaxRects(input, MaxRectsBinPack::FreeRectChoiceHeuristic(0)));
	if(! timer -> TimeLeft()) return;
#ifdef PRINT_LOCAL_INFO
	cerr << "SortDecreasingSideMaxRects(0) Complete: " << answer.GetNumberOfBins() << "\n";
#endif
	answer.CheckMin(solver.MaxRects(input, MaxRectsBinPack::FreeRectChoiceHeuristic(2)));
	if(! timer -> TimeLeft()) return;
#ifdef PRINT_LOCAL_INFO
	cerr << "SortDecreasingSideMaxRects(2) Complete: " << answer.GetNumberOfBins() << "\n";
#endif

	// Global
	answer.CheckMin(solver.MaxRectsGlobal(input, MaxRectsBinPack::FreeRectChoiceHeuristic(0)));
	if(! timer -> TimeLeft()) return;
#ifdef PRINT_LOCAL_INFO
	cerr << "GlobalMaxRectsGlobal(0) Complete: " << answer.GetNumberOfBins() << "\n";
#endif
	answer.CheckMin(solver.MaxRectsGlobal(input, MaxRectsBinPack::FreeRectChoiceHeuristic(4)));
	if(! timer -> TimeLeft()) return;
#ifdef PRINT_LOCAL_INFO
	cerr << "GlobalMaxRectsGlobal(4) Complete: " << answer.GetNumberOfBins() << "\n";
#endif
	answer.CheckMin(solver.GuillotineGlobal(input, true, GuillotineBinPack::FreeRectChoiceHeuristic(0), GuillotineBinPack::GuillotineSplitHeuristic(2)));
	if(! timer -> TimeLeft()) return;
#ifdef PRINT_LOCAL_INFO
	cerr << "GlobalGuillotineGlobal(true,0,2) Complete: " << answer.GetNumberOfBins() << "\n";
	cerr << "Finished All First Phase\n :)\n";
#endif

	/*
	SortDecreasingAreaMaxRects(0): -> Min: 0.364431487 Max: 1.000000000 Mean: 0.671102204 dev: 0.092970704
	SortDecreasingAreaMaxRects(2): -> Min: 0.364431487 Max: 1.000000000 Mean: 0.671030104 dev: 0.092956124
	GlobalMaxRectsGlobal(2): -> Min: 0.364431487 Max: 1.000000000 Mean: 0.671559963 dev: 0.093494635
	SortDecreasingSideMaxRects(0): -> Min: 0.364431487 Max: 1.000000000 Mean: 0.666462138 dev: 0.084676626
	SortDecreasingSideMaxRects(2): -> Min: 0.364431487 Max: 1.000000000 Mean: 0.666449769 dev: 0.084646306
	GlobalMaxRectsGlobal(0): -> Min: 0.364431487 Max: 1.000000000 Mean: 0.665743797 dev: 0.092333672
	GlobalMaxRectsGlobal(4): -> Min: 0.343000000 Max: 1.000000000 Mean: 0.616584215 dev: 0.113505676
	GlobalGuillotineGlobal(true,0,2): -> Min: 0.364431487 Max: 1.000000000 Mean: 0.614500496 dev: 0.108704542
	GlobalGuillotineGlobal(false,0,2): -> Min: 0.364431487 Max: 1.000000000 Mean: 0.612780697 dev: 0.105765537
	*/
}

void _SecondSolvers()
{
	input.SortDecreasingSide();

	answer.CheckMin(solver.Guillotine(input, true, GuillotineBinPack::FreeRectChoiceHeuristic(2), GuillotineBinPack::GuillotineSplitHeuristic(2)));
	if(! timer -> TimeLeft()) return;
#ifdef PRINT_LOCAL_INFO
	cerr << "SortDecreasingSideGuillotine(true,2,2) Complete: " << answer.GetNumberOfBins() << "\n";
#endif
	answer.CheckMin(solver.MaxRects(input, MaxRectsBinPack::FreeRectChoiceHeuristic(3)));
	if(! timer -> TimeLeft()) return;
#ifdef PRINT_LOCAL_INFO
	cerr << "SortDecreasingSideMaxRects(3) Complete: " << answer.GetNumberOfBins() << "\n";
#endif
	answer.CheckMin(solver.MaxRects(input, MaxRectsBinPack::FreeRectChoiceHeuristic(4)));
	if(! timer -> TimeLeft()) return;
#ifdef PRINT_LOCAL_INFO
	cerr << "SortDecreasingSideMaxRects(4) Complete: " << answer.GetNumberOfBins() << "\n";
#endif
	answer.CheckMin(solver.Skyline(input, SkylineBinPack::LevelChoiceHeuristic(1)));
	if(! timer -> TimeLeft()) return;
#ifdef PRINT_LOCAL_INFO
	cerr << "SortDecreasingSideSkyline(WM,1) Complete: " << answer.GetNumberOfBins() << "\n";
#endif
	answer.CheckMin(solver.Skyline(input, SkylineBinPack::LevelChoiceHeuristic(0)));
	if(! timer -> TimeLeft()) return;
#ifdef PRINT_LOCAL_INFO
	cerr << "SortDecreasingSideSkyline(WM,0) Complete: " << answer.GetNumberOfBins() << "\n";
#endif

	// Global
	answer.CheckMin(solver.GuillotineGlobal(input, false, GuillotineBinPack::FreeRectChoiceHeuristic(0), GuillotineBinPack::GuillotineSplitHeuristic(2)));
	if(! timer -> TimeLeft()) return;
#ifdef PRINT_LOCAL_INFO
	cerr << "GlobalGuillotineGlobal(false,0,2) Complete: " << answer.GetNumberOfBins() << "\n";
#endif

	answer.CheckMin(solver.Guillotine(input, true, GuillotineBinPack::FreeRectChoiceHeuristic(2), GuillotineBinPack::GuillotineSplitHeuristic(0)));
	if(! timer -> TimeLeft()) return;
#ifdef PRINT_LOCAL_INFO
	cerr << "SortDecreasingSideGuillotine(true,2,0) Complete: " << answer.GetNumberOfBins() << "\n";
#endif

	// Global
	answer.CheckMin(solver.MaxRectsGlobal(input, MaxRectsBinPack::FreeRectChoiceHeuristic(1)));
	if(! timer -> TimeLeft()) return;
#ifdef PRINT_LOCAL_INFO
	cerr << "GlobalMaxRectsGlobal(1) Complete: " << answer.GetNumberOfBins() << "\n";
#endif
	answer.CheckMin(solver.GuillotineGlobal(input, true, GuillotineBinPack::FreeRectChoiceHeuristic(2), GuillotineBinPack::GuillotineSplitHeuristic(0)));
	if(! timer -> TimeLeft()) return;
#ifdef PRINT_LOCAL_INFO
	cerr << "GlobalGuillotineGlobal(true,2,0) Complete: " << answer.GetNumberOfBins() << "\n";
#endif
	answer.CheckMin(solver.GuillotineGlobal(input, false, GuillotineBinPack::FreeRectChoiceHeuristic(2), GuillotineBinPack::GuillotineSplitHeuristic(0)));
	if(! timer -> TimeLeft()) return;
#ifdef PRINT_LOCAL_INFO
	cerr << "GlobalGuillotineGlobal(false,2,0) Complete: " << answer.GetNumberOfBins() << "\n";
#endif
	answer.CheckMin(solver.GuillotineGlobal(input, false, GuillotineBinPack::FreeRectChoiceHeuristic(1), GuillotineBinPack::GuillotineSplitHeuristic(2)));
	if(! timer -> TimeLeft()) return;
#ifdef PRINT_LOCAL_INFO
	cerr << "GlobalGuillotineGlobal(false,1,2) Complete: " << answer.GetNumberOfBins() << "\n";
#endif


	input.SortDecreasingArea();

	answer.CheckMin(solver.Guillotine(input, true, GuillotineBinPack::FreeRectChoiceHeuristic(1), GuillotineBinPack::GuillotineSplitHeuristic(5)));
	if(! timer -> TimeLeft()) return;
#ifdef PRINT_LOCAL_INFO
	cerr << "SortDecreasingAreaGuillotine(true,1,5) Complete: " << answer.GetNumberOfBins() << "\n";
	cerr << "Finished All Second Phase\n :)\n";
#endif
}

void AllSolvers()
{
	// ShelfBinPack(ALL, WasteMap RectBestShortSideFit SplitMaximizeArea, SortDecreasingArea), O(n^2)
	for(int i = 0; i < 7; i ++)
	{
		if(! timer -> TimeLeft()) return;
		answer.CheckMin(solver.Shelf(input, ShelfBinPack::ShelfChoiceHeuristic(i), GuillotineBinPack::RectBestShortSideFit, GuillotineBinPack::SplitMaximizeArea));
		if(! timer -> TimeLeft()) return;
	}

	// MaxRectsGlobal(ALL), O(n^3)
	for(int i = 0; i < 5; i ++)
	{
		if(! timer -> TimeLeft()) return;
		answer.CheckMin(solver.MaxRects(input, MaxRectsBinPack::FreeRectChoiceHeuristic(i)));
		if(! timer -> TimeLeft()) return;
	}
	
	// SkylineGlobal(ALL), O(n^2)
	for(int i = 0; i < 2; i ++)
	{
		if(! timer -> TimeLeft()) return;
		answer.CheckMin(solver.Skyline(input, SkylineBinPack::LevelChoiceHeuristic(i)));
		if(! timer -> TimeLeft()) return;
	}

	// GuillotineGlobal(ALL, WasteMap), O(n^3)
	for(int i = 0; i < 2; i ++)
	{
		for(int j = 0; j < 3; j ++)
		{
			for(int k = 0; k < 6; k ++)
			{
				if(! timer -> TimeLeft()) return;
				answer.CheckMin(solver.Guillotine(input, i == 0, GuillotineBinPack::FreeRectChoiceHeuristic(j), GuillotineBinPack::GuillotineSplitHeuristic(k)));
				if(! timer -> TimeLeft()) return;
			}
		}
	}
}

void SecondSolvers()
{
	timer -> NextState();
#ifdef PRINT_LOCAL_INFO
	cerr << "\n\nStarting Second Solvers at RunTime: " << timer -> GetElapsedTime() << "\n";
#endif

	for(int iter = 0; ; iter ++)
	{
		if(iter)
		{
		#ifdef PRINT_LOCAL_INFO
			cerr << "RandomShuffle(10)\n";
		#endif
			input.RandomShuffle(10);

			AllSolvers();
		}
		else
		{
			_SecondSolvers();
		}

		/*
		BEST: SortDecreasingSideGuillotine(true,2,2):  -> SCORE: 1147.314026289 MEAN 0.674890604
		BEST: SortDecreasingSideMaxRects(3):  -> SCORE: 1146.512894667 MEAN 0.674419350
		BEST: SortDecreasingSideMaxRects(4):  -> SCORE: 1146.590407759 MEAN 0.674464946
		BEST: SortDecreasingSideSkyline(WM,1):  -> SCORE: 1146.590407759 MEAN 0.674464946
		BEST: SortDecreasingSideSkyline(WM,0):  -> SCORE: 1146.591935853 MEAN 0.674465845
		BEST: GlobalGuillotineGlobal(false,0,2):  -> SCORE: 1147.318228563 MEAN 0.674893076
		BEST: SortDecreasingSideGuillotine(true,2,0):  -> SCORE: 1147.328939326 MEAN 0.674899376
		BEST: GlobalMaxRectsGlobal(1):  -> SCORE: 1147.328939326 MEAN 0.674899376
		BEST: GlobalGuillotineGlobal(true,2,0):  -> SCORE: 1147.365647402 MEAN 0.674920969
		BEST: GlobalGuillotineGlobal(false,2,0):  -> SCORE: 1147.369903811 MEAN 0.674923473
		BEST: GlobalGuillotineGlobal(false,1,2):  -> SCORE: 1147.380624042 MEAN 0.674929779
		BEST: SortDecreasingAreaGuillotine(true,1,5):  -> SCORE: 1147.383316111 MEAN 0.674931362
		*/

		if(! timer -> TimeLeft()) return;
	}
}

int main(array<System::String ^> ^args)
{
	// Dedicated to Momtchil :)
	srand(1366613);

	answer.Init(input);
	
	FirstSolvers();

	SecondSolvers();

	PrintAnswer();
	

    return 0;
}
