#include "stdafx.h"


Solver::Solver(Timer * t)
{
	timer = t;
}

Answer Solver::Shelf(const Input & input, ShelfBinPack::ShelfChoiceHeuristic shelfChoiceHeuristic, GuillotineBinPack::FreeRectChoiceHeuristic freeRectChoiceHeuristic, GuillotineBinPack::GuillotineSplitHeuristic guillotineSplitHeuristic)
{
vector<ShelfBinPack *> binPack;
Answer answer;

	binPack.push_back(new ShelfBinPack(timer, input.GetBoxWidth(), input.GetBoxHeight(), true));
	
	for(int i = 0; i < input.CountRectangles(); ++i)
	{
	bool isUsed = false;
		
		if(! timer -> TimeLeft()) {answer.Init(input); return answer;}

		RectSize rectSize = input.GetRectangle(i);
		
		for(size_t pack = 0; pack < binPack.size(); pack ++)
		{
			if(float(1.0 - binPack[pack] -> Occupancy()) * float(input.GetBoxWidth() * input.GetBoxHeight()) < float(rectSize.width * rectSize.height))
				continue;

			Rect output = binPack[pack] -> Insert(rectSize.width, rectSize.height, shelfChoiceHeuristic, freeRectChoiceHeuristic, guillotineSplitHeuristic);
			
			if(! timer -> TimeLeft()) {answer.Init(input); return answer;}

			if (output.height > 0)
			{
				answer.AddRectangle(output, pack, rectSize.inputIndex);
				
				isUsed = true;
				
				break;
			}
		}
		
		if(! isUsed)
		{
			binPack.push_back(new ShelfBinPack(timer, input.GetBoxWidth(), input.GetBoxHeight(), true));
			
			Rect output = binPack.back() -> Insert(rectSize.width, rectSize.height, shelfChoiceHeuristic, freeRectChoiceHeuristic, guillotineSplitHeuristic);
			
			if(! timer -> TimeLeft()) {answer.Init(input); return answer;}

			if (output.height > 0)
			{
				answer.AddRectangle(output, int(binPack.size()) - 1, rectSize.inputIndex);
			}
		}
	}

	return answer;
}

Answer Solver::Skyline(const Input & input, SkylineBinPack::LevelChoiceHeuristic levelChoiceHeuristic)
{
vector<SkylineBinPack *> binPack;
Answer answer;

	binPack.push_back(new SkylineBinPack(timer, input.GetBoxWidth(), input.GetBoxHeight(), true));
	
	for(int i = 0; i < input.CountRectangles(); ++i)
	{
	bool isUsed = false;
		
		if(! timer -> TimeLeft()) {answer.Init(input); return answer;}

		RectSize rectSize = input.GetRectangle(i);
		
		for(size_t pack = 0; pack < binPack.size(); pack ++)
		{
			if(float(1.0 - binPack[pack] -> Occupancy()) * float(input.GetBoxWidth() * input.GetBoxHeight()) < float(rectSize.width * rectSize.height))
				continue;

			Rect output = binPack[pack] -> Insert(rectSize.width, rectSize.height, levelChoiceHeuristic);
			
			if(! timer -> TimeLeft()) {answer.Init(input); return answer;}

			if (output.height > 0)
			{
				answer.AddRectangle(output, pack, rectSize.inputIndex);
				
				isUsed = true;
				
				break;
			}
		}
		
		if(! isUsed)
		{
			binPack.push_back(new SkylineBinPack(timer, input.GetBoxWidth(), input.GetBoxHeight(), true));
			
			Rect output = binPack.back() -> Insert(rectSize.width, rectSize.height, levelChoiceHeuristic);
			
			if(! timer -> TimeLeft()) {answer.Init(input); return answer;}

			if (output.height > 0)
			{
				answer.AddRectangle(output, int(binPack.size()) - 1, rectSize.inputIndex);
			}
		}
	}

	return answer;
}

Answer Solver::MaxRects(const Input & input, MaxRectsBinPack::FreeRectChoiceHeuristic freeRectChoiceHeuristic)
{
vector<MaxRectsBinPack *> binPack;
Answer answer;

	binPack.push_back(new MaxRectsBinPack(timer, input.GetBoxWidth(), input.GetBoxHeight()));
	
	for(int i = 0; i < input.CountRectangles(); ++i)
	{
	bool isUsed = false;
		
		if(! timer -> TimeLeft()) {answer.Init(input); return answer;}

		RectSize rectSize = input.GetRectangle(i);
		
		for(size_t pack = 0; pack < binPack.size(); pack ++)
		{
			if(float(1.0 - binPack[pack] -> Occupancy()) * float(input.GetBoxWidth() * input.GetBoxHeight()) < float(rectSize.width * rectSize.height))
				continue;

			Rect output = binPack[pack] -> Insert(rectSize.width, rectSize.height, freeRectChoiceHeuristic);
			
			if(! timer -> TimeLeft()) {answer.Init(input); return answer;}

			if (output.height > 0)
			{
				answer.AddRectangle(output, pack, rectSize.inputIndex);
				
				isUsed = true;
				
				break;
			}
		}
		
		if(! isUsed)
		{
			binPack.push_back(new MaxRectsBinPack(timer, input.GetBoxWidth(), input.GetBoxHeight()));
			
			Rect output = binPack.back() -> Insert(rectSize.width, rectSize.height, freeRectChoiceHeuristic);
			
			if(! timer -> TimeLeft()) {answer.Init(input); return answer;}

			if (output.height > 0)
			{
				answer.AddRectangle(output, int(binPack.size()) - 1, rectSize.inputIndex);
			}
		}
	}

	return answer;
}

Answer Solver::Guillotine(const Input & input, bool merge, GuillotineBinPack::FreeRectChoiceHeuristic freeRectChoiceHeuristic, GuillotineBinPack::GuillotineSplitHeuristic guillotineSplitHeuristic)
{
vector<GuillotineBinPack *> binPack;
Answer answer;


	binPack.push_back(new GuillotineBinPack(timer, input.GetBoxWidth(), input.GetBoxHeight()));
	
	for(int i = 0; i < input.CountRectangles(); ++i)
	{
	bool isUsed = false;
		
		if(! timer -> TimeLeft()) {answer.Init(input); return answer;}	

		RectSize rectSize = input.GetRectangle(i);
		
		for(size_t pack = 0; pack < binPack.size(); pack ++)
		{
			if(float(1.0 - binPack[pack] -> Occupancy()) * float(input.GetBoxWidth() * input.GetBoxHeight()) < float(rectSize.width * rectSize.height))
				continue;

			Rect output = binPack[pack] -> Insert(rectSize.width, rectSize.height, merge, freeRectChoiceHeuristic, guillotineSplitHeuristic);
			
			if(! timer -> TimeLeft()) {answer.Init(input); return answer;}

			if (output.height > 0)
			{
				answer.AddRectangle(output, pack, rectSize.inputIndex);
				
				isUsed = true;
				
				break;
			}
		}
		
		if(! isUsed)
		{
			binPack.push_back(new GuillotineBinPack(timer, input.GetBoxWidth(), input.GetBoxHeight()));
			
			Rect output = binPack.back() -> Insert(rectSize.width, rectSize.height, merge, freeRectChoiceHeuristic, guillotineSplitHeuristic);
			
			if(! timer -> TimeLeft()) {answer.Init(input); return answer;}

			if (output.height > 0)
			{
				answer.AddRectangle(output, int(binPack.size()) - 1, rectSize.inputIndex);
			}
		}
	}

	return answer;
}

Answer Solver::MaxRectsGlobal(const Input & input, MaxRectsBinPack::FreeRectChoiceHeuristic freeRectChoiceHeuristic)
{
MaxRectsBinPack * binPack = new MaxRectsBinPack(timer, input.GetBoxWidth(), input.GetBoxHeight());
vector<RectSize> rectangles;
vector<Rect> binOutput;
Answer answer;
	
	if(! timer -> TimeLeft())
	{answer.Init(input); return answer;}

	for(int i = 0; i < input.CountRectangles(); ++i)
	{
		rectangles.push_back(input.GetRectangle(i));
	}
	
	for(int binIndex = 0; rectangles.size(); binIndex ++)
	{
		binOutput.clear();
		binPack -> Init(timer, input.GetBoxWidth(), input.GetBoxHeight());

		binPack -> Insert(rectangles, binOutput, freeRectChoiceHeuristic);

		if(! timer -> TimeLeft()) {answer.Init(input); return answer;}

		for(size_t i = 0; i < binOutput.size(); i ++)
		{
			answer.AddRectangle(binOutput[i], binIndex, binOutput[i].inputIndex);
		}
	}

	return answer;
}

Answer Solver::SkylineGlobal(const Input & input, SkylineBinPack::LevelChoiceHeuristic levelChoiceHeuristic)
{
SkylineBinPack * binPack = new SkylineBinPack(timer, input.GetBoxWidth(), input.GetBoxHeight(), true);
vector<RectSize> rectangles;
vector<Rect> binOutput;
Answer answer;
	
	if(! timer -> TimeLeft()) {answer.Init(input); return answer;}

	for(int i = 0; i < input.CountRectangles(); ++i)
	{
		rectangles.push_back(input.GetRectangle(i));
	}
	
	for(int binIndex = 0; rectangles.size(); binIndex ++)
	{
		binOutput.clear();
		binPack -> Init(timer, input.GetBoxWidth(), input.GetBoxHeight(), true);

		binPack -> Insert(rectangles, binOutput, levelChoiceHeuristic);

		if(! timer -> TimeLeft()) {answer.Init(input); return answer;}

		for(size_t i = 0; i < binOutput.size(); i ++)
		{
			answer.AddRectangle(binOutput[i], binIndex, binOutput[i].inputIndex);
		}
	}

	return answer;
}

Answer Solver::GuillotineGlobal(const Input & input, bool merge, GuillotineBinPack::FreeRectChoiceHeuristic freeRectChoiceHeuristic, GuillotineBinPack::GuillotineSplitHeuristic guillotineSplitHeuristic)
{
GuillotineBinPack * binPack = new GuillotineBinPack(timer, input.GetBoxWidth(), input.GetBoxHeight());
vector<RectSize> rectangles;
vector<Rect> binOutput;
Answer answer;
	
	if(! timer -> TimeLeft()) {answer.Init(input); return answer;}

	for(int i = 0; i < input.CountRectangles(); ++i)
	{
		rectangles.push_back(input.GetRectangle(i));
	}
	
	for(int binIndex = 0; rectangles.size(); binIndex ++)
	{
		binOutput.clear();
		binPack -> Init(timer, input.GetBoxWidth(), input.GetBoxHeight());

		binPack -> Insert(rectangles, binOutput, merge, freeRectChoiceHeuristic, guillotineSplitHeuristic);

		if(! timer -> TimeLeft()) {answer.Init(input); return answer;}

		for(size_t i = 0; i < binOutput.size(); i ++)
		{
			answer.AddRectangle(binOutput[i], binIndex, binOutput[i].inputIndex);
		}
	}
	
	return answer;
}